#! logic_tree.py

"""
Classes to define logic tree structures
"""

from copy import deepcopy
from dataclasses import dataclass, field
from functools import reduce
from itertools import product
from math import isclose
from operator import add, mul
from typing import Any, Dict, Generator, Iterable, List, Union


@dataclass
class BranchAttributeSpec:
    name: str
    long_name: str
    value_options: List[Any] = field(default_factory=list)


@dataclass(frozen=True)
class BranchAttributeValue:
    name: str
    long_name: str
    value: Any = None

    @staticmethod
    def from_branch_attribute(ba: BranchAttributeSpec, value):
        return BranchAttributeValue(ba.name, ba.long_name, value)

    @staticmethod
    def all_from_branch_attribute(ba: BranchAttributeSpec):
        for opt in ba.value_options:
            yield BranchAttributeValue(ba.name, ba.long_name, opt)

    def __repr__(self):
        return f"{self.name}{self.value}"


@dataclass
class Branch:
    values: List[BranchAttributeValue]
    weight: float = 1.0
    onfault_nrml_id: Union[str, None] = ""
    distributed_nrml_id: Union[str, None] = ""
    inversion_solution_id: Union[str, None] = ""
    inversion_solution_type: Union[str, None] = ""
    rupture_set_id: Union[str, None] = ""


@dataclass
class FaultSystemLogicTreeSpec:
    short_name: str
    long_name: str
    branches: List['BranchAttributeValue'] = field(default_factory=list)


@dataclass
class FaultSystemLogicTree:
    short_name: str
    long_name: str
    branches: List['Branch'] = field(default_factory=list)

    def validate_weights(self) -> bool:
        weight = 0.0
        for b in self.branches:
            weight += b.weight
        return weight == 1.0

    def derive_spec(self) -> FaultSystemLogicTreeSpec:
        fslt_spec = FaultSystemLogicTreeSpec(short_name=self.short_name, long_name=self.long_name)

        options: Dict[str, set] = {}

        # extract unique values in to options
        for branch in self.branches:
            # iterate all the branches, yielding the unique  branch options
            # print('branch', branch)
            for value in branch.values:
                if value.name not in options.keys():
                    options[value.name] = set([])
                val = value.value
                if isinstance(val, list):  # make it hashable
                    val = tuple(val)
                bao = BranchAttributeValue(value.name, value.long_name, val)
                options[value.name].add(bao)

        # print(options)

        def option_values(options: Dict) -> Generator:
            # boil down the options values
            ret = {}
            for key, opt in options.items():
                if key not in ret:
                    ret[key] = BranchAttributeSpec(key, list(opt)[0].long_name)

                values = []
                for val in opt:
                    values.append(val.value)
                ret[key].value_options = sorted(values)

            for key, bas in ret.items():
                yield bas

        fslt_spec.branches = list(option_values(options))

        return fslt_spec


@dataclass
class SourceLogicTreeCorrelation:
    primary_short_name: str
    secondary_short_name: str
    primary_values: List[BranchAttributeValue]
    secondary_values: List[BranchAttributeValue]

    # these methods enforce set compairson
    def is_primary(self, bavs: Iterable[BranchAttributeValue]) -> bool:
        return set(self.primary_values) == set(bavs)

    def is_secondary(self, bavs: Iterable[BranchAttributeValue]) -> bool:
        return set(self.secondary_values) == set(bavs)


@dataclass
class SourceLogicTreeSpec:
    fault_system_lts: List[FaultSystemLogicTreeSpec] = field(default_factory=list)


@dataclass
class SourceLogicTree:
    version: str
    title: str
    fault_system_lts: List[FaultSystemLogicTree] = field(default_factory=list)
    correlations: List[SourceLogicTreeCorrelation] = field(
        default_factory=list
    )  # to use for selecting branches and re-weighting when logic trees are correlated

    def derive_spec(self) -> SourceLogicTreeSpec:
        slt_spec = SourceLogicTreeSpec()
        for fslt in self.fault_system_lts:
            slt_spec.fault_system_lts.append(FaultSystemLogicTree.derive_spec(fslt))
        return slt_spec


@dataclass
class CompositeBranch:
    """Combination of all fault type branches"""

    branches: List[Branch]
    weight: float = 1.0

    def __post_init__(self) -> None:
        self.weight = reduce(mul, [branch.weight for branch in self.branches])


@dataclass
class FlattenedSourceLogicTree:
    """Flattened source logic tree containing all CompositeBranch combinations"""

    version: str
    title: str
    branches: List[CompositeBranch]

    def __post_init__(self) -> None:
        total_weight = reduce(add, [branch.weight for branch in self.branches])
        if not isclose(total_weight, 1.0):
            raise Exception('logic tree weights do not add to 1.0 (sum is %s)' % total_weight)

    @classmethod
    def from_source_logic_tree(cls, slt: SourceLogicTree):

        slt_copy = deepcopy(slt)  # don't want to change weights of origional logic tree object
        branches = [fslt.branches for fslt in slt_copy.fault_system_lts]

        def yield_cor(branches, slt_copy):
            nnames = [[faultsys_lt.short_name] * len(faultsys_lt.branches) for faultsys_lt in slt_copy.fault_system_lts]
            for cb, names in zip(product(*branches), product(*nnames)):
                has_correlation = False
                for cor in slt_copy.correlations:
                    sindex = names.index(cor.secondary_short_name)
                    if cor.is_secondary(cb[sindex].values):
                        has_correlation = True
                        pindex = names.index(cor.primary_short_name)
                        if cor.is_primary(cb[pindex].values):
                            cb[sindex].weight = 1.0
                            yield CompositeBranch(list(cb))
                if not has_correlation:
                    yield CompositeBranch(list(cb))

        if slt.correlations:
            return cls(slt.version, slt.title, list(yield_cor(branches, slt_copy)))
        else:
            return cls(slt.version, slt.title, [CompositeBranch(list(cb)) for cb in product(*branches)])

    def __repr__(self):
        return f'{self.__class__} title {self.title} number of branches: {len(self.branches)}'



#! nshm_v1_0_0.py

import itertools
import json
from pathlib import Path

import dacite  # for dataclass reconstitution

import nzshm_model.source_logic_tree.SLT_v8_gmm_v2_final as slt_config  # NOQA F401
from nzshm_model.source_logic_tree.logic_tree import (
    Branch,
    BranchAttributeSpec,
    BranchAttributeValue,
    FaultSystemLogicTree,
    SourceLogicTree,
)

version = 'NSHM_v1.0.0'
title = "Initial version"

json_slt = Path(__file__).parent / "source_logic_tree" / "nshm_v1.0.0.json"


def source_logic_tree():
    return dacite.from_dict(data_class=SourceLogicTree, data=json.load(open(json_slt)))


def build_crustal_branches():
    # define the branch attributes for our FaultSystemLogicTree
    C = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='C', long_name='area-magnitude scaling', value_options=[4])
    )
    s = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='s', long_name='moment rate scaling', value_options=[0.5, 1.0])
    )
    bN = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='bN', long_name='bN pair', value_options=[(0.87, 25), (0.95, 19.2)])
    )
    dm = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='dm', long_name='deformation model', value_options=['Geodetic', 'Geologic'])
    )

    crustal_branches = FaultSystemLogicTree('Cru', 'Crustal')
    for (a, b, c, d) in itertools.product(C, s, bN, dm):
        crustal_branches.branches.append(Branch(values=[a, b, c, d], weight=0.125, onfault_nrml_id='ABC'))
    return crustal_branches



#! python toshi_api.py

"""
enrich SLT from TOSHI_API.

"""

import base64
import dataclasses
import json
import os
from typing import Dict, List, Optional, Union

import boto3
from botocore.exceptions import ClientError
from nshm_toshi_client.toshi_client_base import ToshiClientBase

API_URL = os.getenv('NZSHM22_TOSHI_API_URL', "http://127.0.0.1:5000/graphql")


# TODO this should be define in an AWS helper library  - it's used everywhere
def get_secret(secret_name, region_name):
    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(service_name='secretsmanager', region_name=region_name)

    # In this sample we only handle the specific exceptions for the 'GetSecretValue' API.
    # See https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
    # We rethrow the exception by default.

    try:
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'DecryptionFailureException':
            # Secrets Manager can't decrypt the protected secret text using the provided KMS key.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'InternalServiceErrorException':
            # An error occurred on the server side.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'InvalidParameterException':
            # You provided an invalid value for a parameter.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'InvalidRequestException':
            # You provided a parameter value that is not valid for the current state of the resource.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'ResourceNotFoundException':
            # We can't find the resource that you asked for.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
    else:
        # Decrypts secret using the associated KMS CMK.
        # Depending on whether the secret is a string or binary, one of these fields will be populated.
        if 'SecretString' in get_secret_value_response:
            return json.loads(get_secret_value_response['SecretString'])
        else:
            return base64.b64decode(get_secret_value_response['SecretBinary'])


@dataclasses.dataclass
class InversionInfo:
    typename: Union[str, None] = None
    solution_id: Union[str, None] = None


class SourceSolutionMap:
    """A mapping between nrml ids and hazard solution ids"""

    def __init__(self, hazard_jobs: List[dict] = []) -> None:
        self._dict: Dict[str, str] = {}
        if hazard_jobs:
            for job in hazard_jobs:
                for arg in job['node']['child']['arguments']:
                    if arg['k'] == 'logic_tree_permutations':
                        branch_info = json.loads(arg['v'].replace("'", '"'))[0]['permute'][0]['members'][0]
                        onfault_nrml_id = branch_info['inv_id']
                        distributed_nrml_id = branch_info['bg_id']
                hazard_solution = job['node']['child']['hazard_solution']
                self._dict[self.__key(onfault_nrml_id, distributed_nrml_id)] = hazard_solution['id']

    def append(self, other: 'SourceSolutionMap'):
        self._dict.update(other._dict)

    def get_solution_id(self, *, onfault_nrml_id: str, distributed_nrml_id: str) -> Optional[str]:
        return self._dict.get(self.__key(onfault_nrml_id, distributed_nrml_id))

    @staticmethod
    def __key(onfault_nrml_id: str, distributed_nrml_id: str) -> str:
        return ':'.join((str(onfault_nrml_id), str(distributed_nrml_id)))


class ToshiApi(ToshiClientBase):
    def get_source_from_nrml(self, nrml_id):
        qry = '''
        query nrml ($nrml_id: ID!) {
          node(id: $nrml_id) {
            __typename
            ... on InversionSolutionNrml {
              file_name
              source_solution {
                __typename
                ... on Node { id }
              }
            }
          }
        }'''

        # print(qry)
        input_variables = dict(nrml_id=nrml_id)
        executed = self.run_query(qry, input_variables)
        return (
            InversionInfo(
                typename=executed['node']['source_solution']['__typename'],
                solution_id=executed['node']['source_solution']['id'],
            )
            if executed.get('node')
            else InversionInfo()
        )

    def get_rupture_set_id(self, solution_id):
        qry = '''
        query file0 ($id: ID!) {
          node(id: $id) {
            __typename
            ... on PredecessorsInterface {
              predecessors {
                id
                typename
                relationship
                depth
                file_node: node {
                  __typename
                  ... on Node {id}
                  ... on File {
                    file_name
                  }
                }
              }
            }
          }
        }
        '''
        # print(qry)
        input_variables = dict(id=solution_id)
        # print(input_variables)
        executed = self.run_query(qry, input_variables)
        return executed['node']


if 'TEST' in API_URL.upper():
    API_KEY = get_secret("NZSHM22_TOSHI_API_SECRET_TEST", "us-east-1").get("NZSHM22_TOSHI_API_KEY_TEST")
elif 'PROD' in API_URL.upper():
    API_KEY = get_secret("NZSHM22_TOSHI_API_SECRET_PROD", "us-east-1").get("NZSHM22_TOSHI_API_KEY_PROD")
else:
    API_KEY = os.getenv('NZSHM22_TOSHI_API_KEY', "")
headers = {"x-api-key": API_KEY}
toshi_api = ToshiApi(API_URL, None, with_schema_validation=False, headers=headers)


def solution_rupt_set_id(solution_id) -> str:
    data = toshi_api.get_rupture_set_id(solution_id)
    for itm in data['predecessors']:
        if itm["typename"] == "File":
            return itm['file_node']['id']
    return ""


if __name__ == "__main__":
    """This should be a test!"""
    import dataclasses
    from pathlib import Path

    from nzshm_model.source_logic_tree.slt_config import from_config

    SKIP_FS_NAMES = ['SLAB', 'HIK']  # , 'CRU'

    config_path = Path(__file__).parent / 'SLT_v8_gmm_v2_final.py'
    slt = from_config(config_path)

    for fslt in slt.fault_system_lts:
        if fslt.short_name in SKIP_FS_NAMES:  # CRU
            continue
        for branch in fslt.branches[-2:]:
            nrml_info = toshi_api.get_source_from_nrml(branch.onfault_nrml_id)
            print(nrml_info)
            branch.inversion_solution_id = nrml_info.solution_id
            branch.inversion_solution_type = nrml_info.typename
            branch.rupture_set_id = solution_rupt_set_id(nrml_info.solution_id)
            print(branch)  # , end='', flush=True)

    j = json.dumps(dataclasses.asdict(slt), indent=4)
    # print(j)



#! test_module_import
import dataclasses
import json
from pathlib import Path

import dacite
import pytest

import nzshm_model
from nzshm_model.source_logic_tree.logic_tree import Branch, FaultSystemLogicTree, SourceLogicTree
from nzshm_model.source_logic_tree.slt_config import (
    decompose_crustal_tag,
    decompose_subduction_tag,
    from_config,
    get_config_group,
    get_config_group_tag_permutations,
    get_config_groups,
)


class TestStructure:
    def setup(self):
        self.model = nzshm_model.get_model_version('NSHM_v1.0.0')

    def test_slt_groups_v1_0_0(self):
        v1_0_0 = self.model
        groups = list(get_config_groups(v1_0_0.slt_config.logic_tree_permutations))
        for slt in groups:
            print(f'group {slt}')
        assert len(groups) == 4

    def test_slt_get_group(self):
        group = get_config_group(self.model.slt_config.logic_tree_permutations, 'HIK')
        print(group)
        assert group['group'] == 'Hik'

    def test_get_config_group_tag_permutations(self):
        group = list(get_config_group_tag_permutations(self.model.slt_config.logic_tree_permutations, 'HIK'))
        print(group)
        assert group[0] == 'Hik TL, N16.5, b0.95, C4, s0.42'

    def test_decompose_crustal_tag_permutations(self):
        group = list(get_config_group_tag_permutations(self.model.slt_config.logic_tree_permutations, 'CRU'))
        print(group)
        print(list(decompose_crustal_tag(group[0])))
        assert len(group) == 36

    def test_decompose_hik_tag_permutations(self):
        group = list(get_config_group_tag_permutations(self.model.slt_config.logic_tree_permutations, 'HIK'))
        print(group)
        print(list(decompose_subduction_tag(group[0])))
        assert len(group) == 9

    def test_decompose_puy_tag_permutations(self):
        group = list(get_config_group_tag_permutations(self.model.slt_config.logic_tree_permutations, 'PUY'))
        print(group)
        print(list(decompose_subduction_tag(group[0])))
        assert len(group) == 3

    def test_assemble_puy(self):

        group_key = 'PUY'
        group = get_config_group(self.model.slt_config.logic_tree_permutations, group_key)
        print(group)

        fslt = FaultSystemLogicTree('PUY', 'Puysegur')

        for member in group['members']:
            fslt.branches.append(
                Branch(
                    values=list(decompose_subduction_tag(member['tag'])),
                    weight=member['weight'],
                    onfault_nrml_id=member['inv_id'],
                    distributed_nrml_id=member['bg_id'],
                )
            )

        print(fslt)

        assert fslt.branches[-1].values[0].value == "0.7"
        assert fslt.branches[-1].values[1].value == [0.902, 4.6]
        assert fslt.branches[-1].values[2].value == 4.0
        assert fslt.branches[-1].values[3].value == 1.72


class TestConfigSerialisation:
    @pytest.mark.parametrize("python_module", ['SLT_v9p0p0.py', 'SLT_v8_gmm_v2_final.py'])
    def test_slt_dict_to_json(self, python_module):
        config = Path(__file__).parent.parent / 'nzshm_model' / 'source_logic_tree' / python_module
        slt = from_config(config)
        obj = dataclasses.asdict(slt)
        jsonish = json.dumps(obj, indent=2)
        assert obj
        assert jsonish

    @pytest.mark.parametrize("python_module", ['SLT_v9p0p0.py', 'SLT_v8_gmm_v2_final.py'])
    def test_slt_v8_round_trip(self, python_module):

        config = Path(__file__).parent.parent / 'nzshm_model' / 'source_logic_tree' / python_module
        slt = from_config(config)
        obj = dataclasses.asdict(slt)
        jsonish = json.dumps(obj, indent=2)

        new_slt_0 = dacite.from_dict(data_class=SourceLogicTree, data=obj)
        assert slt == new_slt_0

        new_slt_1 = dacite.from_dict(data_class=SourceLogicTree, data=json.loads(jsonish))
        print("SLT")
        print(slt.fault_system_lts[0])
        print("SLT from dict")
        print(new_slt_0.fault_system_lts[0])

        print("SLT from json")
        print(new_slt_1.fault_system_lts[0])

        assert slt == new_slt_1


class TestFromConfig:
    def test_slt_v8(self):
        config = Path(__file__).parent.parent / 'nzshm_model' / 'source_logic_tree' / 'SLT_v8_gmm_v2_final.py'
        slt = from_config(config)
        print(slt)
        assert slt.fault_system_lts[0].branches[-1].values[0].name == 'dm'

    def test_large_SLT_example_A_crustal(self):
        config = Path(__file__).parent / 'fixtures' / 'large_SLT_example_A.py'
        slt = from_config(config)
        print(slt)
        assert slt.fault_system_lts[1].branches[-1].values[0].name == 'dm'
        assert slt.fault_system_lts[1].branches[-1].values[0].value == 'TC'


class TestDecomposeTags:
    def test_large_SLT_example_A_crustal(self):
        decomp = list(decompose_crustal_tag("Cru_geol, b0.849, C4.1, s0.53"))
        print(decomp[0].value)

        assert decomp[0].name == 'dm'
        assert decomp[0].value == 'geologic'
        assert decomp[1].name == 'b'
        assert decomp[1].value == 0.849
        assert decomp[2].name == 'C'
        assert decomp[2].value == 4.1
        assert decomp[3].name == 's'
        assert decomp[3].value == 0.53

    def test_large_SLT_V8_crustal(self):
        decomp = list(decompose_crustal_tag("geodetic, TI, N2.7, b0.823 C4.2 s0.66"))
        assert decomp[0].name == 'dm'
        assert decomp[0].value == 'geodetic'
        assert decomp[1].name == 'td'
        assert decomp[1].value is False
        assert decomp[2].name == 'bN'
        assert decomp[2].value == [0.823, 2.7]
        assert decomp[3].name == 'C'
        assert decomp[3].value == 4.2
        assert decomp[4].name == 's'
        assert decomp[4].value == 0.66



#! nshm_v1_0_4.py

import json
from pathlib import Path

import dacite  # for dataclass reconstitution

import nzshm_model.source_logic_tree.SLT_v9p0p0 as slt_config  # NOQA F401
from nzshm_model.source_logic_tree.logic_tree import SourceLogicTree

version = 'NSHM_v1.0.4'
title = "NSHM version 1.0.4, corrected fault geometry"

json_slt = Path(__file__).parent / "source_logic_tree" / "nshm_v1.0.4.json"


def source_logic_tree():
    return dacite.from_dict(data_class=SourceLogicTree, data=json.load(open(json_slt)))
